/* cups-config-daemon.c - Configure cups printer drivers via D-BUS
 *
 * Copyright (C) 2002, 2003  Red Hat Inc.
 *
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include <dbus/dbus.h>
#include <dbus/dbus-glib.h>
#include <dbus/dbus-glib-lowlevel.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <signal.h>


#include "config.h"
#include "cups-config-daemon.h"
#include "rb-debug.h"

G_DEFINE_TYPE(CupsConfigDaemon, cups_config_daemon, G_TYPE_OBJECT)

static void cups_config_daemon_dispose  (GObject *object);
static void cups_config_daemon_finalize (GObject *object);

static DBusHandlerResult cups_config_daemon_dbus_message_handler (DBusConnection *connection, 
                                                                  DBusMessage *message, 
                                                                  void *user_data);
static void cups_config_daemon_dbus_unregister_handler           (DBusConnection *connection, 
                                                                  void *user_data);

static int idle_check_dbus                  (CupsConfigDaemon *confd);
static void cups_config_daemon_set_driver   (CupsConfigDaemon *confd, DBusMessage *message);
static void cups_config_daemon_match_driver (CupsConfigDaemon *confd, DBusMessage *message);

struct CupsConfigDaemonPrivate
{
	DBusConnection *dbus;
	guint idle_check_dbus_id;

	gboolean disposed;
};

static GObjectClass *parent_class = NULL;

static void
cups_config_daemon_class_init (CupsConfigDaemonClass *klass)
{
	GObjectClass *object_class;
	object_class = G_OBJECT_CLASS (klass);

	rb_debug ("cups_config_daemon_class_init");

	parent_class = g_type_class_peek_parent (klass);

	object_class->dispose = cups_config_daemon_dispose;
	object_class->finalize = cups_config_daemon_finalize;
}

void
cups_config_daemon_init (CupsConfigDaemon *confd)
{
	confd->priv = g_new0 (struct CupsConfigDaemonPrivate, 1);

	rb_debug ("cups_config_daemon_init");

	confd->priv->dbus = NULL;
	confd->priv->idle_check_dbus_id = g_idle_add ((GSourceFunc) idle_check_dbus,
						       confd);
}

static void
cups_config_daemon_dispose (GObject *object)
{
	CupsConfigDaemon *confd;
	confd = CUPS_CONFIG_DAEMON (object);

	rb_debug ("cups_config_daemon_dispose");

	if (confd->priv->disposed)
		return;
	confd->priv->disposed = TRUE;

	if (confd->priv->dbus) {
		dbus_connection_unref (confd->priv->dbus);
		confd->priv->dbus = NULL;
	}

	if (confd->priv->idle_check_dbus_id) {
		g_source_remove (confd->priv->idle_check_dbus_id);
		confd->priv->idle_check_dbus_id = 0;
	}

}

static void
cups_config_daemon_finalize (GObject *object)
{
	CupsConfigDaemon *confd;
	confd = CUPS_CONFIG_DAEMON (object);

	rb_debug ("cups_config_daemon_dispose");

	g_return_if_fail (confd->priv != NULL);

	g_free (confd->priv);
}
       
CupsConfigDaemon *
cups_config_daemon_new (void)
{
	rb_debug ("cups_config_daemon_new");
        return CUPS_CONFIG_DAEMON (g_object_new (CUPS_TYPE_CONFIG_DAEMON, NULL));
}

static int
idle_check_dbus (CupsConfigDaemon *confd)
{
	DBusConnection *dbus;
	DBusError error;
	DBusObjectPathVTable cups_config_daemon_vtable = { &cups_config_daemon_dbus_unregister_handler, 
                                                           &cups_config_daemon_dbus_message_handler, 
                                                           NULL, NULL, NULL, NULL};
	rb_debug ("idle_check_dbus");

	dbus_error_init (&error);
	dbus = dbus_bus_get (DBUS_BUS_SYSTEM, &error);
	if (dbus == NULL) {
		g_warning ("DbusConfigDaemon: %s", error.message);
		dbus_error_free (&error);

		confd->priv->idle_check_dbus_id = g_timeout_add (5000,
								(GSourceFunc) idle_check_dbus,
								confd);
		return FALSE;
	}

	dbus_bus_request_name (dbus, CUPS_DRIVER_CONFIG_NAMESPACE, 0, &error);
	if (dbus_error_is_set (&error)) {
		g_error ("Could not acquire service %s.  dbus_bus_acquire_service() says: '%s'\n", CUPS_DRIVER_CONFIG_NAMESPACE, error.message);

		dbus_error_free (&error);
		return FALSE;
	}

	if (!dbus_connection_register_object_path (dbus, CUPS_DRIVER_CONFIG_PATH, &cups_config_daemon_vtable, confd)) {
		g_error ("Could not register a handler for %s\n", CUPS_DRIVER_CONFIG_PATH);
		return FALSE;
	}

	dbus_connection_setup_with_g_main (dbus, NULL);
	confd->priv->dbus = dbus;

	return FALSE;

}

static DBusHandlerResult 
cups_config_daemon_dbus_message_handler (DBusConnection *connection, 
                                         DBusMessage *message, 
                                         void *user_data)
{
	const char *method;
	const char *path;
	CupsConfigDaemon *confd;
	DBusHandlerResult res;

	rb_debug ("cups_config_daemon_dbus_message_handler");

	res = DBUS_HANDLER_RESULT_NOT_YET_HANDLED;

	confd = CUPS_CONFIG_DAEMON (user_data);

	method = dbus_message_get_member (message);
	path = dbus_message_get_path (message);

	if (strcmp (CUPS_DRIVER_CONFIG_SET_DRIVER_METHOD, method) == 0) {
		cups_config_daemon_set_driver (confd, message);
		res = DBUS_HANDLER_RESULT_HANDLED;
	} else if (strcmp (CUPS_DRIVER_CONFIG_MATCH_DRIVER_METHOD, method) == 0) {
		cups_config_daemon_match_driver (confd, message);
		res = DBUS_HANDLER_RESULT_HANDLED;
	} 
 
	return (res);
}


static void 
cups_config_daemon_dbus_unregister_handler (DBusConnection *connection, void *user_data)
{
	/* do nothing */
}

/*
* Some makes in the Foomatic database are in abreviated form
* and must be translated back to a more readable name
*/
static gchar *
translate_make (gchar *make)
{
	gchar *up_make;
        gchar *result;

	up_make = g_ascii_strup (make, -1);
	if (g_ascii_strcasecmp (up_make, "HEWLETT-PACKARD") == 0)
		result = g_strdup ("Hewlett-Packard");
	else
		result = g_strdup (make);

	g_free (up_make);
	return result;
}

/* change an occurance of character from to character to in string s */
void 
strchrep(char *s, char from, char to) 
{ 
      while ((s = strchr(s, from)) != NULL)
        *s = to;
}

static void
cups_config_daemon_match_driver (CupsConfigDaemon *confd, DBusMessage *message)
{
	DBusError error;
	DBusMessage *return_message;
	DBusMessageIter iter;
	const gchar *make, *model;
	gchar *sane_make;
	GString *match_driver_cmd;
	gchar *driver;

	dbus_error_init (&error);	

	if (!dbus_message_get_args (message, &error,
                                    DBUS_TYPE_STRING, &make,
                                    DBUS_TYPE_STRING, &model,
                                    DBUS_TYPE_INVALID))
	{
		g_warning ("Error getting message aguments %s\n", error.message);

		return_message = dbus_message_new_error (message, error.name, error.message);
		dbus_error_free (&error);

		goto out;
	}
	
	sane_make = translate_make (make);

	strchrep (sane_make, '\"', '_');
	strchrep (model, '\"', '_');

	match_driver_cmd = g_string_new (PRINTCONF_EXEC);
	g_string_append (match_driver_cmd, " --Xmatch-driver --make=\"");
	g_string_append (match_driver_cmd, sane_make);
	g_string_append (match_driver_cmd, "\" --model=\"");
	g_string_append (match_driver_cmd, model);
	g_string_append (match_driver_cmd, "\"");

	rb_debug ("Running: %s", match_driver_cmd->str);

	g_spawn_command_line_sync (match_driver_cmd->str, &driver, NULL, NULL, NULL);

	if (driver != NULL && g_ascii_strcasecmp(driver, "")) {
		return_message = dbus_message_new_method_return (message);
		dbus_message_iter_init_append (return_message, &iter);
		dbus_message_iter_append_basic (&iter, DBUS_TYPE_STRING, &driver);
	} else {
		return_message = dbus_message_new_error_printf (message, CUPS_DRIVER_CONFIG_DRIVER_NOT_FOUND_ERROR, "A driver for %s %s could not be found.", make, model);
	} 
	
	g_free (sane_make);
	g_free (driver);
	g_string_free (match_driver_cmd, TRUE);
	
	out:
		dbus_connection_send (confd->priv->dbus, return_message, NULL);
	  	dbus_connection_flush (confd->priv->dbus);
  		dbus_message_unref (return_message);
}

char *
readlink_malloc (const char *filename)
{
  int size = 100;
  char *buffer = NULL;

  while (1)
  {
    buffer = (char *) realloc (buffer, size);
    int nchars = readlink (filename, buffer, size);
    if (nchars < 0)
    {
      free (buffer);
      return NULL;
    }
    if (nchars < size)
      return buffer;
    size *= 2;
  }
}

static void
cups_config_daemon_set_driver (CupsConfigDaemon *confd, DBusMessage *message)
{
	DBusError error;
	const gchar *make, *model, *printer_udi, *printer_name;
	int vendor_len, model_len, cmd_len, name_len, device_len;
	int add_len;
	char vendor_switch[] = " --make=";
	char model_switch[] = " --model=";
	char name_switch[] = " --name=";
	char device_switch[] = " --device=";
	char device_proto[] = "hal://"; 
	char add_cmd[] = " --Xadd-local";
	char *printconf;
	char *sane_make;

	gchar *cups_pid_buf;
	int cups_pid_len;
	GError *cups_pid_error;

	dbus_error_init (&error);	

	if (!dbus_message_get_args (message, &error,
                                    DBUS_TYPE_STRING, &make,
                                    DBUS_TYPE_STRING, &model,
                                    DBUS_TYPE_STRING, &printer_udi,
                                    DBUS_TYPE_STRING, &printer_name,
                                    DBUS_TYPE_INVALID))
	{
		g_warning ("Error getting message aguments %s\n", error.message);
		return;
	}

	rb_debug ("cups_config_daemon_set_driver - Make: %s, Model: %s, Printer UDI: %s, Printer Name: %s", 
 		   make, model, printer_udi, printer_name);

	sane_make = translate_make (make);

	vendor_len = strlen (sane_make) + strlen (vendor_switch) + 2;
	model_len = strlen (model) + strlen (model_switch) + 2;
	name_len = strlen (printer_name) + strlen (name_switch) + 2;
	device_len = strlen (printer_udi) + strlen (device_switch) + strlen (device_proto) + 2;

	/*add command length*/
	cmd_len = strlen (add_cmd);
	add_len = vendor_len + name_len + model_len + cmd_len + device_len + 1;

  	/* make sure no one tries to unquote untrusted strings */
	strchrep (sane_make, '\"', '_');
	strchrep (model, '\"', '_');
	strchrep (model, '\\', '_');
	strchrep (sane_make, '\\', '_');

	printconf = malloc (sizeof(char) * (add_len + strlen (PRINTCONF_EXEC)));

	if (printconf == NULL) 
		g_error ("Out of memory error. Exiting!");

	sprintf (printconf, "%s%s%s\"%s\"%s\"%s\"%s\"%s\"%s\"%s%s\"", PRINTCONF_EXEC, 
                                                                     add_cmd, 
                                                                     vendor_switch, sane_make, 
                                                                     model_switch, model,
                                                                     name_switch, printer_name,
						                     device_switch, device_proto, printer_udi);

	system (printconf);
	system ("/usr/sbin/printconf-backend --force-rebuild");

	if (g_file_get_contents ("/var/run/cupsd.pid",
                                 &cups_pid_buf,
                                 &cups_pid_len,
                                 &cups_pid_error))
	{
		GPid cups_pid;

		cups_pid = atoi (cups_pid_buf);

		rb_debug ("Cups pid: %s", cups_pid_buf);
		if (cups_pid > 0) {
			GString *proc_pid_file_name;
			gchar *pids_exe_name;

			proc_pid_file_name = g_string_new ("");
			g_string_printf (proc_pid_file_name, "/proc/%s/exe", cups_pid_buf);

			pids_exe_name = readlink_malloc (proc_pid_file_name->str);
			g_string_free (proc_pid_file_name, TRUE);

			if (pids_exe_name != NULL) {
				if (strcmp (pids_exe_name, "/usr/sbin/cupsd") == 0)
					kill (cups_pid, SIGHUP);
				else
					system ("/usr/bin/killall -SIGHUP cupsd");

      				g_free (pids_exe_name);
    			} else {
				system ("/usr/bin/killall -SIGHUP cupsd");
			}
  		}

		g_free (cups_pid_buf);
	} else {
		system ("/usr/bin/killall -SIGHUP cupsd");
		g_error_free (cups_pid_error);
	}



	g_free (printconf); 
	g_free (sane_make);
}
