/*
 *
 *   a halafied local printer configuration app based on lpadmin from cups
 *
 *   Copyright 1997-2003 by Easy Software Products, all rights reserved.
 *   Copyright (C) 2004 Red Hat, Inc.
 *
 *   These coded instructions, statements, and computer programs are the
 *   property of Easy Software Products and are protected by Federal
 *   copyright law.  Distribution and use rights are outlined in the file
 *   "LICENSE" which should have been included with this file.  If this
 *   file is missing or damaged please contact Easy Software Products
 *   at:
 *
 *       Attn: CUPS Licensing Information
 *       Easy Software Products
 *       44141 Airport View Drive, Suite 204
 *       Hollywood, Maryland 20636-3111 USA
 *
 *       Voice: (301) 373-9603
 *       EMail: cups-info@cups.org
 *         WWW: http://www.cups.org
 *
 *   This file is subject to the Apple OS-Developed Software exception.
 *
 * Contents:
 *
 *   main()         - Send a file to the specified HAL device.
 *   list_devices() - List all HAL printer devices.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <cups/cups.h>
#include <cups/language.h>
#include <config.h>
#ifdef HAVE_LIBZ
#  include <zlib.h>
#endif /* HAVE_LIBZ */
#include <libhal.h>
#include <limits.h>
#include <sys/types.h>
#include <signal.h>
#include <unistd.h>

#define URI_PREFIX "hal://"

/*
 * The debug macros are used if you compile with DEBUG defined.
 *
 * Usage:
 *
 *   DEBUG_puts("string")
 *   DEBUG_printf(("format string", arg, arg, ...));
 *
 * Note the extra parenthesis around the DEBUG_printf macro...
 */


/* Helper macros */

#  ifdef DEBUG
#    define DEBUG_puts(x) puts(x)
#    define DEBUG_printf(x) printf x
#  else
#    define DEBUG_puts(x)
#    define DEBUG_printf(x)
#  endif /* DEBUG */

/* change an occurance of character from to character to in string s */
void 
strchrep(char *s, char from, char to) 
{ 
      while ((s = strchr(s, from)) != NULL)
        *s = to;
}

static int	validate_name(char *);

static void	remove_printer_using_printconf(char *, char *, char *, char *);
static int	set_printer_driver_using_printconf(char *, char *, char *, char *, dbus_bool_t);
static int	printer_needs_config(char *, char *, char *, char *, DBusError *);

static void	print_usage_and_exit();
static void	add_hal_device_if_printer(LibHalContext *, char *);
static void	remove_hal_device_if_printer(LibHalContext *, char *);

static char 	   *construct_printer_name(char *, int);
static dbus_bool_t dbus_get_user_driver (DBusConnection *, char *, char *, char *, char *, char **, char **, DBusError *);
static void        dbus_prompt_printer_driver (DBusConnection *, char *, char *, char *, char *, DBusError *);

static void	sig_hup_cups();

static LibHalContext *hal_init(void);

/*
 * 'hal_init()' - Initializes a HAL context
 */

LibHalContext *
hal_init(void)
{
	LibHalContext *hal_ctx;
	DBusError error;
	DBusConnection *dbus_conn;

	hal_ctx = libhal_ctx_new ();

	if (hal_ctx == NULL) {
		fprintf (stderr, "ERROR: Unable to access HAL daemon\n");
		return NULL;
	}

	dbus_error_init (&error);
	dbus_conn = dbus_bus_get (DBUS_BUS_SYSTEM, &error);

	if (dbus_error_is_set (&error)) {
		fprintf (stderr, "ERROR: Could not connect to system bus: %s\n", error.message);
		dbus_error_free (&error);
		libhal_ctx_free (hal_ctx);
		return NULL;
	}

	libhal_ctx_set_dbus_connection (hal_ctx, dbus_conn);

	if (!libhal_ctx_init (hal_ctx, &error)) {
		fprintf (stderr, "ERROR: Could initialize HAL daemon: %s\n", error.message);
		dbus_error_free (&error);
		libhal_ctx_free (hal_ctx);
		return NULL;
	}

	return hal_ctx;
}


/*
 * 'main()' - Parse options and configure the scheduler.
 */

int
main(int  argc,			/* I - Number of command-line arguments */
     char *argv[])		/* I - Command-line arguments */
{

  char		*hal_udi;	/* UDI of hal device*/  
  int		i;		/* Looping var */
  int		remove_device;  /* if true then we are removing the device, else adding */
  LibHalContext *hal_ctx;       /* Hal context */
  char **capabilities;
  int is_printer;

  hal_udi     = NULL;
  remove_device = 0;

  for (i = 1; i < argc; i ++)
    if (argv[i][0] == '-')
      switch (argv[i][1])
      {
        case 'p' : /* device UID */
            if (argv[i][2])
            {
              hal_udi = argv[i] + 2;
            }
            else
            { 
               i ++;

	      if (i >= argc)
                hal_udi = NULL;
              else
                hal_udi = argv[i];
            }
            break;

        case 'x' : /* remove device */
            remove_device = 1;

            if (argv[i][2])
            {
              hal_udi = argv[i] + 2;
            }
            else
            { 
               i ++;

	      if (i >= argc)
                hal_udi = NULL;
              else
                hal_udi = argv[i];
            }
            break;
      }
  
  if (hal_udi == NULL) {
    hal_udi = getenv("UDI");
    if (hal_udi == NULL)
       print_usage_and_exit();
  }

  /* get hal context */
  if ((hal_ctx = hal_init()) == NULL) 
    exit(1);

  capabilities = libhal_device_get_property_strlist(hal_ctx, hal_udi, "info.capabilities", NULL);
 
  is_printer = 0;

  if (capabilities != NULL) {
    for (i=0; capabilities[i] != NULL && is_printer == 0; i++) {
      if (strcmp(capabilities[i], "printer") == 0)
        is_printer = 1;
    }
    
    libhal_free_string_array (capabilities);
  }
   
  if (!is_printer)
    return (-1);

  if (!remove_device)  
    add_hal_device_if_printer(hal_ctx, hal_udi);
  else
    remove_hal_device_if_printer(hal_ctx, hal_udi);

  return (0);
}

/*
 * 'print_usage_and_exit()' - prints out the command line usage and exits
 */

static void
print_usage_and_exit()
{
  fprintf(stderr, "USAGE: hal_lpadmin -p <hal device udi> [-x]");
  exit(1);
}

/*
 * 'add_hal_device_if_printer()' - checks if the device is a printer and adds it to cups
 */
static void
add_hal_device_if_printer(LibHalContext *ctx, 
                          char *udi)
{
  char *product;
  char *vendor;
  char *description;
  char *serial;
  char *device;
  int usb_port_num;
  char *uri = NULL;
  char *printer = NULL;

  product = libhal_device_get_property_string(ctx, udi, "printer.product", NULL);
  if (product == NULL)
  {
    product = libhal_device_get_property_string(ctx, udi, "usb.product", NULL);
    if (product == NULL)
      product = strdup("Unknown");
    else
      product = strdup(product);
  }
  else
    product = strdup(product);

  vendor = libhal_device_get_property_string(ctx, udi, "printer.vendor", NULL);
  if (vendor == NULL)
  {
    vendor = libhal_device_get_property_string(ctx, udi, "usb.vendor", NULL);
    if (vendor == NULL)
      vendor = strdup("Unknown");
    else
      vendor = strdup(vendor);
  }
  else
    vendor = strdup(vendor);

  description = libhal_device_get_property_string(ctx, udi, "printer.description", NULL);
  serial = libhal_device_get_property_string(ctx, udi, "printer.serial", NULL);
  device = libhal_device_get_property_string(ctx, udi, "printer.device", NULL);
  usb_port_num = libhal_device_get_property_int(ctx, udi, "usb.port_number", NULL);

  uri = malloc(sizeof(char) * (strlen(udi) + strlen(URI_PREFIX) + 1));

  if (!uri) 
  {
    free(vendor);
    free(product);
    fprintf(stderr, "ERROR: Out of memory\n"); 
    return;
  }

  sprintf(uri, "%s%s", URI_PREFIX, udi); 
  
  printer = construct_printer_name(product, usb_port_num);

  if (!printer) { 
    free(vendor);
    free(product);
    free(uri);
    return;
  }

  set_printer_driver_using_printconf(printer, vendor, product, udi, TRUE);

  free(vendor);
  free(product);
  free(uri);
}

/*
 * 'remove_hal_device_if_printer()' - checks if the device is a printer and removes it from cups
 */
static void
remove_hal_device_if_printer(LibHalContext *ctx, 
                             char *udi)
{
  char *product;
  char *serial;
  char *vendor;
  int usb_port_num;
  char *printer = NULL;

  product = libhal_device_get_property_string(ctx, udi, "printer.product", NULL);
  if (product == NULL)
  {
    product = libhal_device_get_property_string(ctx, udi, "usb.product", NULL);
    if (product == NULL)
      product = strdup("Unknown");
    else
      product = strdup(product);
  } 
  else
    product = strdup(product);

  vendor = libhal_device_get_property_string(ctx, udi, "printer.vendor", NULL);
  if (vendor == NULL)
  {
    vendor = libhal_device_get_property_string(ctx, udi, "usb.vendor", NULL);
    if (vendor == NULL)
      vendor = strdup("Unknown");
    else
      vendor = strdup(vendor);
  }
  else
    vendor = strdup(vendor);

  serial = libhal_device_get_property_string(ctx, udi, "printer.serial", NULL);
  usb_port_num = libhal_device_get_property_int(ctx, udi, "usb.port_number", NULL);

  printer = construct_printer_name(product, usb_port_num); 

  if (!printer)
  {
    free(product);
    free(vendor);
    return;
  }

  remove_printer_using_printconf(printer, vendor, product, udi);

  free(product);
  free(vendor);
  free(printer);
}

/*
 * 'validate_name()' - Make sure the printer name only contains valid chars,
 *                     replace spaces with dashes and make all letters lowercase.
 */

static int			/* O - 0 if name is no good, 1 if name is good */
validate_name(char *name)	/* I - Name to check */
{
  char	*ptr;		/* Pointer into name */


 /*
  * Scan the whole name...
  */

  for (ptr = name; *ptr; ptr ++)
    if (*ptr == '@')
      break;
    else if (*ptr == ' ')
      *ptr = '-';
    else if ((*ptr >= 0 && *ptr < ' ') || *ptr == 127 || *ptr == '/')
      return (0);
    else
      *ptr = tolower(*ptr);

 /*
  * All the characters are good; validate the length, too...
  */

  return ((ptr - name) < 128);
}

/*
 * 'construct_printer_name()' -  creates a name by taking the product name
 *                               and appending it with the usb port number 
 *                    
 */
static char *
construct_printer_name(char *product, int usb_port_num)
{
  char *printer;
  char port_num[256];

  snprintf(port_num, 256, "-%i", usb_port_num);
  printer = malloc(sizeof(char) * (strlen(port_num) + strlen(product) + 1)); 

  if (!printer) 
  {
    fprintf(stderr, "ERROR: Out of memory\n"); 
    return NULL;
  }

  sprintf(printer, "%s%s", product, port_num);

  if (!validate_name(printer)) 
  {
    fprintf(stderr, "Printer name %s, contains non printable characters\n", printer); 
    free(printer);
    return NULL;
  }
 
  return printer; 
}

static char *
translate_make(char *make)
{
  char *uc_make;
  int make_len, i;
  char *result;

  make_len = strlen(make);
  uc_make = strdup(make);

  for(i=0; i<make_len; i++)
  {
    uc_make[i] = toupper(uc_make[i]);
  }

  if (strcmp(uc_make, "HEWLETT-PACKARD") == 0)
    result = strdup("HP");
  else
    result = strdup(make);

  free (uc_make);
  return result;
}

char *
readlink_malloc (char *filename)
{
  int size = 100;
  char *buffer = NULL;

  while (1)
  {
    buffer = (char *) realloc (buffer, size);
    int nchars = readlink (filename, buffer, size);
    if (nchars < 0)
    {
      free (buffer);
      return NULL;
    }
    if (nchars < size)
      return buffer;
    size *= 2;
  }
}

static void
sig_hup_cups ()
{
  FILE *cups_pid_file;
  char cups_pid_buf[256];
  int cups_pid;

  cups_pid_file = fopen ("/var/run/cupsd.pid", "r");

  if (cups_pid_file != NULL) 
  {
    fgets (cups_pid_buf, sizeof (cups_pid_buf), cups_pid_file);
    cups_pid = atoi (cups_pid_buf);
    fclose (cups_pid_file);
  } else
    goto fail;

  if (cups_pid > 0)
  {
    char proc_pid_file_name[265];
    char *pids_exe_name;

    snprintf (proc_pid_file_name, 265, "/proc/%s/exe", cups_pid_buf);

    pids_exe_name = readlink_malloc (proc_pid_file_name);

    if (pids_exe_name != NULL) 
    {
      if (strcmp (pids_exe_name, "/usr/sbin/cupsd") == 0)
      {
        kill (cups_pid, SIGHUP);
        return;
      }
      free (pids_exe_name);
    }
  }

  fail:
    system ("/usr/bin/killall -SIGHUP cupsd");
}

/*TODO: This is all just hacked together
        eventualy we want to break out the
        printconf stuff and make it saner*/

void
remove_printer_using_printconf(char *printer, 
                               char *vendor, 
                               char *model,
                               char *udi)
{
  int vendor_len, model_len, cmd_len, device_len;
  int remove_len;
  char vendor_switch[] = " --make=";
  char model_switch[] = " --model=";
  char device_switch[] = " --device=";
  char device_proto[] = "hal://"; 
  char remove_cmd[] = " --Xremove-local";
  char *make;
  char *printconf;

  make = translate_make(vendor);

  vendor_len = strlen(make) + strlen(vendor_switch) + 2;
  model_len = strlen(model) + strlen(model_switch) + 2;
  device_len = strlen(udi) + strlen(device_switch) + strlen(device_proto) + 2;

  /*remove command length*/
  cmd_len = strlen(remove_cmd);
  remove_len = vendor_len + model_len + cmd_len + device_len + 1;

  /* make sure no one tries to unquote untrusted strings */
  strchrep(vendor, '\"', '_');
  strchrep(model, '\"', '_');
  strchrep(model, '\\', '_');
  strchrep(make, '\\', '_');

  printconf = malloc(sizeof(char) * (remove_len + strlen(PRINTCONF_EXEC)));

  if (printconf == NULL) 
  {
    fprintf(stderr, "Out of memory error. Exiting!");
    exit(-1);
  }

  printf ("s: %i, l: %i\n", remove_len + strlen(PRINTCONF_EXEC), strlen(printconf));
  
  sprintf(printconf, "%s%s%s\"%s\"%s\"%s\"%s\"%s%s\"", PRINTCONF_EXEC, 
                                                       remove_cmd, 
                                                       vendor_switch, make, 
                                                       model_switch, model,
						       device_switch, device_proto, udi);

  system(printconf);
  printf (printconf);
  system("/usr/sbin/printconf-backend --force-rebuild");

  sig_hup_cups ();

  free(printconf);
  free(make);
}



static int
set_printer_driver_using_printconf(char *printer, 
                                   char *vendor, 
                                   char *model,
                                   char *udi,
                                   dbus_bool_t prompt_session_for_driver)
{
  FILE *driver_pipe;
  char driver[PATH_MAX + 1];
  size_t chars_read;
  int vendor_len, model_len, cmd_len, name_len, device_len;
  int match_len, add_len;
  char vendor_switch[] = " --make=";
  char model_switch[] = " --model=";
  char name_switch[] = " --name=";
  char device_switch[] = " --device=";
  char device_proto[] = "hal://"; 
  char match_cmd[] = " --Xmatch-driver";
  char add_cmd[] = " --Xadd-local";
  char *printconf;
  char *make;
  int result;

  make = translate_make(vendor);

  vendor_len = strlen(make) + strlen(vendor_switch) + 2;
  model_len = strlen(model) + strlen(model_switch) + 2;
  name_len = strlen(printer) + strlen(name_switch) + 2;
  device_len = strlen(udi) + strlen(device_switch) + strlen(device_proto) + 2;

  /*match command length*/
  cmd_len = strlen(match_cmd);
  match_len = vendor_len + model_len + cmd_len + 1;

  /*add command length*/
  cmd_len = strlen(add_cmd);
  add_len = vendor_len + name_len + model_len + cmd_len + device_len + 1;


  /* make sure no one tries to unquote untrusted strings */
  strchrep(make, '\"', '_');
  strchrep(model, '\"', '_');
  strchrep(model, '\\', '_');
  strchrep(make, '\\', '_');

  printconf = malloc(sizeof(char) * (match_len + strlen(PRINTCONF_EXEC)));

  if (printconf == NULL) 
  {
    fprintf(stderr, "Out of memory error. Exiting!");
    exit(-1);
  }

  sprintf(printconf, "%s%s%s\"%s\"%s\"%s\"", PRINTCONF_EXEC, 
                                             match_cmd, 
                                             vendor_switch, make, 
                                             model_switch, model);

  driver_pipe = popen(printconf, "r");

  chars_read = fread(driver, sizeof(char), PATH_MAX, driver_pipe);
  driver[chars_read] = '\0';

  result = pclose(driver_pipe);

  free(printconf);

  if (result == 0)
  {
    printconf = malloc(sizeof(char) * (add_len + strlen(PRINTCONF_EXEC)));

    if (printconf == NULL) 
    {
      fprintf(stderr, "Out of memory error. Exiting!");
      exit(-1);
    }

    sprintf(printconf, "%s%s%s\"%s\"%s\"%s\"%s\"%s\"%s\"%s%s\"", PRINTCONF_EXEC, 
                                                       add_cmd, 
                                                       vendor_switch, make, 
                                                       model_switch, model,
                                                       name_switch, printer,
						       device_switch, device_proto, udi);

    system(printconf);
    system("/usr/sbin/printconf-backend --force-rebuild");

    sig_hup_cups ();
    free(printconf); 
  }
  else if (prompt_session_for_driver)
  {
    DBusError error;

    dbus_error_init(&error);
    if (printer_needs_config(vendor, model, udi, printer, &error) != 0) 
    {
      if (dbus_error_is_set (&error))
      {
        fprintf (stderr, "An error occurred while calling printer_needs_config: %s\n", error.message);
        dbus_error_free (&error);
      }	
    }
  }

  free(make);
  return result;

}

static void
dbus_prompt_printer_driver (DBusConnection *connection,
                            char *make,
                            char *model,
                            char *udi,
                            char *printer,
                            DBusError *error)
{
  DBusMessage *message;
  DBusMessageIter iter;

  message = dbus_message_new_method_call ("com.redhat.PrintDriverSelection",
                                          "/com/redhat/PrintDriverSelection",
                                          "com.redhat.PrintDriverSelection",
                                          "PromptPrintDriver");
                                                                                                                             
  dbus_message_iter_init_append (message, &iter);
  dbus_message_iter_append_basic (&iter, DBUS_TYPE_STRING, &make);         /*make*/
  dbus_message_iter_append_basic (&iter, DBUS_TYPE_STRING, &model);        /*model*/
  dbus_message_iter_append_basic (&iter, DBUS_TYPE_STRING, &udi);          /*printer udi*/
  dbus_message_iter_append_basic (&iter, DBUS_TYPE_STRING, &printer);      /*printer name*/
                                                                                                                             
  dbus_connection_send (connection, message, NULL);
  dbus_connection_flush (connection);
  dbus_message_unref (message);
}

dbus_bool_t 
dbus_get_user_driver (DBusConnection *connection,
                      char *make,
                      char *model,
                      char *udi,
                      char *printer,
                      char **return_make,
                      char **return_model,
                      DBusError *error)
{
  DBusMessage *message;
  DBusMessageIter iter;
  DBusMessage *reply_message;
  char *rmake;
  char *rmodel;

  message = dbus_message_new_method_call ("com.redhat.PrintDriverSelection",
                                          "/com/redhat/PrintDriverSelection",
                                          "com.redhat.PrintDriverSelection",
                                          "GetUserDriver");
                                                                                                                             
  dbus_message_iter_init_append (message, &iter);
  dbus_message_iter_append_basic (&iter, DBUS_TYPE_STRING, &make);         /*make*/
  dbus_message_iter_append_basic (&iter, DBUS_TYPE_STRING, &model);        /*model*/
  dbus_message_iter_append_basic (&iter, DBUS_TYPE_STRING, &udi);          /*printer udi*/
  dbus_message_iter_append_basic (&iter, DBUS_TYPE_STRING, &printer);      /*printer name*/
                                                                                                                             
  reply_message = dbus_connection_send_with_reply_and_block (connection, message, 10000, error);

  if (dbus_error_is_set (error))
    return FALSE;

  if (!dbus_message_get_args (reply_message, error,
                              DBUS_TYPE_STRING, &rmake, 
                              DBUS_TYPE_STRING, &rmodel,
                              DBUS_TYPE_INVALID))
  {
    *return_make = NULL;
    *return_model = NULL;
    dbus_error_free (error);
    return FALSE;
  }

  *return_make = strdup (rmake);
  *return_model = strdup (rmodel);

  dbus_connection_flush (connection);
  dbus_message_unref (reply_message);
  dbus_message_unref (message);

  return TRUE;
}


static int
printer_needs_config(char *make, 
                     char *model, 
                     char *udi, 
                     char *printer, 
                     DBusError *error)
{
  DBusConnection *connection;
  dbus_bool_t result;
  char *return_make;
  char *return_model;

  connection = dbus_bus_get (DBUS_BUS_SYSTEM, error);
  if (connection == NULL) 
  {
    return -1;
  }

  result = dbus_get_user_driver (connection, make, model, udi, printer, &return_make, &return_model, error);

  if (!result)
  {
    if (dbus_error_is_set (error))
      return -1;

    dbus_prompt_printer_driver (connection, make, model, udi, printer, error);
  }
  else
  {
    set_printer_driver_using_printconf (printer, return_make, return_model, udi, FALSE);

    if (return_make != NULL)
      free (return_make);
    if (return_model != NULL)
      free (return_model);
  }

  if (dbus_error_is_set (error))
      return -1;

  dbus_connection_unref (connection);

  return 0;
}

